
"""$Id"""
__version__='$Revision: 1.1 $'[11:-2]

from Shared.DC.ZRDB.THUNK import THUNKED_TM
from DateTime import DateTime
from string import strip, split
from thread import get_ident

import pywintypes, pythoncom, win32com.client
from win32com.client import gencache

gencache.EnsureModule('{00025E01-0000-0000-C000-000000000046}', 0, 4, 0)
win32com.client.Dispatch('DAO.DBEngine.35')
constants = win32com.client.constants

types = {
    constants.dbBoolean: 'i',
    constants.dbByte: 'i',
    constants.dbInteger: 'i',
    constants.dbLong: 'i',
    constants.dbCurrency: 'n',
    constants.dbSingle: 'n',
    constants.dbDouble: 'n',
    constants.dbDate: 'd',
    constants.dbText: 's',
    constants.dbLongBinary: 's',
    constants.dbMemo: 's',
    constants.dbGUID: 's',
}

database_error = pywintypes.com_error

# init DAO engines on a per-thread basis
_e = {}
def get_engine():
    thread_id = get_ident()
    try: 
        return _e[thread_id]
    except KeyError:
        pythoncom.CoInitialize()
        _e[thread_id] = e = win32com.client.Dispatch('DAO.DBEngine.35')
        return e

# init databases on a per-thread basis
_db = {}
def get_db(conn):
    thread_id = get_ident()
    try: 
        return _db[thread_id]
    except KeyError:
        e = get_engine()
        try:
            _db[thread_id] = db = e.OpenDatabase(conn)
        except database_error, exc:
            raise 'Connect Error', exc[2][2]
        return db

def close_dbs():
    for id, db in _db.items():
        db.Close()
        del _db[id]

# NOTE: _begin, _finish, _abort, and query 
# must always be called from the same thread 
# during a transaction or this will not work

class DB(THUNKED_TM):

    database_error = database_error
    opened = ''
    
    def __init__(self, connection):
        """Construct the DB object."""
        self.connection = connection
        self.open()
            
    def open(self):
        """Open the database."""
        get_db(self.connection)
        self.opened=DateTime()

    def close(self):
        """Close the database."""
        close_dbs()
        del self.opened
        
    def _begin(self):
        """Begin a transaction."""
        engine = get_engine()
        engine.Workspaces[0].BeginTrans()

    def _finish(self):
        """Commit a transaction."""
        engine = get_engine()
        engine.Workspaces[0].CommitTrans()

    def _abort(self):
        """Rollback a transaction."""
        engine = get_engine()
        engine.Workspaces[0].Rollback()

    def query(self, query_string, max_rows=99999999):
        """Execute a query."""
        self._register()
        db = get_db(self.connection)
        queries=filter(None, map(strip,split(query_string, '\0')))
        if not queries: raise 'Query Error', 'empty query'
        result = []
        try:
            for q in queries:
                query = db.CreateQueryDef('', q)
                try: recordset = query.OpenRecordset()
                except self.database_error, exc:
                    if (exc[0] == 0x800a0c93): recordset = None
                    else: raise
                if (recordset is not None):
                    fields = query.Fields
                    nrow = 0
                    while (nrow < max_rows) and (not recordset.EOF):
                        record = []
                        for field in recordset.Fields:
                            record.append(field.Value)
                        result.append(record)
                        nrow = nrow + 1
                        recordset.MoveNext()
                else:
                    query.Execute()
        except self.database_error, exc:
            raise 'Query Error', exc[2][2]
            
        if (not result):
            return (), ()

        items = []
        for field in fields:
            items.append({
                'name': field.Name,
                'type': types[field.Type],
                'width': field.Size,
                'null': not field.Required,
                })
                
        return items, result
    
